#include <stdio.h>
#include <errno.h>
#include <unistd.h>	/* fcntl */
#include <fcntl.h>	/* fcntl */
#include <sys/types.h>	/* fichero */
#include <sys/stat.h>	/* fichero */

#define TRUE 1
#define FALSE 0

int fd;		/* descriptor del fichero */
char nombreBD[50];	/* nombre de la base de datos */
int abierta;		/* indica si hay una base de datos abierta */


/* registros de la base de datos */
typedef struct TIPO_REGISTRO{
  int id;
  char nombre[30];
  char apellido[30];
}REGISTRO;

REGISTRO reg;


/* funcin que muestra el n del error, el mensaje asociado y finaliza la ejecucin*/
void Error(void) {

  fprintf(stderr, "ERROR %d. ", errno);
  fprintf(stderr, "%s\n", strerror(errno));
  exit(1);
}

/***************************************************************************/
/* Funcin que muestra el menu en pantalla y devuelve un numero con la opcion
seleccionada */
int Menu(void) {
int opcion;

  printf("\n");
  printf("------------------------------------\n");
  printf("1. Crear base de datos.\n");
  printf("2. Abrir base de datos.\n");
  printf("3. Buscar registro.\n");
  printf("4. Insertar registro.\n");
  printf("5. Modificar registro.\n");
  printf("6. Borrar registro.\n");
  printf("7. Cerrar base de datos.\n");
  printf("0. Salir\n\n");

  printf("Opcion: ");
  scanf("%d",&opcion);

  return(opcion);
}

/***************************************************************************/
/* Funcion que crea un base de datos. El file descriptor se guarda en la 
variable global fd.  */
void Crear(void) {

  printf("Introduce el nombre de la base de datos: ");
  scanf("%s", &nombreBD);

  fd = creat(nombreBD, S_IREAD | S_IWRITE);
  if (fd == -1) 
    Error();

  abierta = TRUE;
}

/***************************************************************************/
/* Procedimiento que abre una base de datos existente */
void Abrir(void) {

  if (abierta) { 
    close(fd);
    printf("\nCerrando...\n");
    abierta = FALSE;
  }

  printf("Introduce el nombre de la base de datos a abrir: ");
  scanf("%s", &nombreBD);
 
  fd = open(nombreBD, O_RDWR);
  if (fd == -1) 
    Error();

  abierta = TRUE;

}


/***************************************************************************
 Muestra por pantalla el contenido del registro. 			   
***************************************************************************/
void Mostrar(void) {

  printf("Nombre: %s\n", reg.nombre);
  printf("Apellido: %s\n", reg.apellido);
  printf("Identificador: %d\n", reg.id);
}


/***************************************************************************
 Inserta un nuevo registro al final del fichero. Introduce todos los     
 datos en un buffer y luego los escribe en el fichero.                   
***************************************************************************/
void Insertar(void) {
char ch;
struct flock lock;
int aux;

  
  if (!abierta) {
    printf("\nBase de datos  no abierta.\n");
    Abrir();
  }
  lseek(fd, 0L, SEEK_END);		/* posicionarse al final */

  /* entrada de datos */
  printf("Nombre: ");
  scanf("%30s", reg.nombre);
  printf("Apellido: ");
  scanf("%30s", reg.apellido);
  printf("Identificador: ");
  scanf("%d", &reg.id);

  /* bloqueo */
  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_CUR;
  lock.l_start = 0;
  lock.l_len = sizeof(REGISTRO);
  if (fcntl(fd, F_SETLKW, &lock) == -1)
    Error();
  
  /* escritura en fichero del buffer de datos */
  if (write(fd, &reg, sizeof(REGISTRO)) == 0)
    Error();
  else {
    printf("Entra un numero para continuar\n");
    scanf("%d", &aux);	/* pausa */
  }

  /* desbloqueo */
  lock.l_type = F_UNLCK; 
  lock.l_start = (-1) * sizeof(REGISTRO);
  if (fcntl(fd, F_SETLKW, &lock) == -1)
    Error(); 

}


/***************************************************************************
 Busqueda secuencial de un identificador en el fichero. En el peor de los 
 casos se recorre todo el fichero. Se devuelve la posicin del registro en 
 el fichero o -1 si no se encuentra.
***************************************************************************/
int Buscar(void) {
int aux;    		/* identificador a buscar */
int encontrado = FALSE;
int fin = FALSE;
struct flock lock;

  if (!abierta) {
    printf("\nBase no abierta.\n");
    Abrir();
  }  

  printf("\nIntroduce el identificador a buscar: ");
  scanf("%d", &aux);
  lseek(fd, 0L, SEEK_SET);		/* posicionar al principio */

  /* se compara hasta encontrarlo o llegar al final del fichero */
  while (!fin && !encontrado) {
    /* bloqueo */
    lock.l_type = F_RDLCK;
    lock.l_whence = SEEK_CUR;
    lock.l_start = 0;
    lock.l_len = sizeof(REGISTRO);
    if (fcntl(fd, F_SETLKW, &lock) == -1)
      Error();
    
    if (read(fd, &reg, sizeof(REGISTRO)) == 0)
      fin = TRUE;	/* final del fichero */
    else
      if (aux == reg.id) { 
        encontrado = TRUE; /* registro encontrado */
        printf("Entra un numero para continuar\n");
        scanf("%d", &aux); /* pausa */
      }
    /* desbloqueo */
    lock.l_type = F_UNLCK; 
    lock.l_start = (-1) * sizeof(REGISTRO);
    if (fcntl(fd, F_SETLKW, &lock) == -1)
      Error(); 
  }

  if (encontrado) {
    Mostrar();
    /* devolvemos la posicion actual */
    return (lseek(fd, 0L, SEEK_CUR) - sizeof(REGISTRO));
  }
  else
   return -1;
}


/***************************************************************************
  Procedimiento que cierra la base de datos actual
***************************************************************************/
void Cerrar(void) {
   
  if (abierta)
    close(fd);
}

/***************************************************************************/
/* Procedimiento que borra un registro. Ponemos un -1 en su identificador */
void Borrar(void) {
int aux;
struct flock lock;  

  aux = Buscar();
  if (aux == -1)
    printf("\nEl registro no existe.\n");
  else {
    reg.id = -1;
    lseek(fd, aux, SEEK_SET);
    /* bloqueo */
    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_CUR;
    lock.l_start = 0;
    lock.l_len = sizeof(REGISTRO);
    if (fcntl(fd, F_SETLKW, &lock) == -1)
      Error();

    write(fd, &reg, sizeof(REGISTRO));

    /* desbloqueo */
    lock.l_type = F_UNLCK; 
    lock.l_start = (-1) * sizeof(REGISTRO);
    if (fcntl(fd, F_SETLKW, &lock) == -1)
      Error(); 

    printf("\nRegistro borrado.");
  }
  
}

/***************************************************************************/
int main(int argc, char *argv[]){
int opcion;

  do {
    opcion = Menu();
    switch (opcion) {
      case 1: Crear();
	      break;
      case 2: Abrir();
	      break;
      case 3: if (Buscar() == -1)
    		printf("\nEl registro no existe.\n");
	      break;
      case 4: Insertar();
	      break;
      case 5: break;
      case 6: Borrar();
	      break;
      case 7: Cerrar();
              break;
 
     default: break;
    }
  }while (opcion != 0);

  Cerrar();
  return 0;
}
