#include "analizador.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#define MAXFILES 20	/* n maximo de archivos */
#define MAXCHAR 50	/* longitud mxima de nombre de fichero */
#define MAXARG 20	/* n mximo de argumentos */
#define NIL -2		/* para fd que no se usan en ese momento*/		

#define PROMPT "[OscarSH]$ "

/******************************************** TIPOS *********************************/
typedef enum {FALSE = 0, TRUE = 1} boolean;

/******************************************** VARIABLES *********************************/
extern char *yytext;	/* parte de la cadena de entrada correspondiente a un token */
extern int yyleng;

/****************************************** PROTOTIPOS *********************************/
void Pausa(void);
void Manejador();
void Ignorar_sig(void);
void Activar_sig(void);
void Esperar(int pid);
void Redireccionar(int fuente_fd, char *fuente, int destino_fd, char *destino, boolean anadir, boolean bckgrnd);
int Invocar(int n_arg, char *argumentos[], int fuente_fd, char *fuente, int destino_fd, char *destino, boolean anadir, boolean bckgrnd);
int Comando(int *pid, boolean tuberia, int *pipefd);

/************************************************************************
Funcion que realiza una pausa en espera de que se pulse ENTER.
************************************************************************/
void Pausa(void) {

  printf("Presione ENTER para continuar...");
  getchar();
}

/********************************************************************************************************
Funcin que hace que se ignoren las seales SIGINT y SIGQUIT.
*********************************************************************************************************/
void Manejador() {
static boolean primera = TRUE;

  if (primera) {
    primera = FALSE;
    printf("\nSeal de fin recibida.Matando a todos los hijos.Pulse Ctrl+C para terminar.\n");
    /* kill(0, SIGTERM); */
  }
  else
    exit(0);
}

/********************************************************************************************************
Funcin que hace que se ignoren las seales SIGINT y SIGQUIT.
*********************************************************************************************************/
void Ignorar_sig(void) {

  signal(SIGINT, SIG_IGN);
  signal(SIGQUIT, SIG_IGN);
}

/********************************************************************************************************
Funcin que activa a la operacion por defecto las seales SIGINT y SIGQUIT.
*********************************************************************************************************/
void Activar_sig(void) {

  signal(SIGINT, SIG_DFL);
  signal(SIGQUIT, SIG_DFL);
}
/********************************************************************************************************
Funcin que espera la finalizacin de procesos hasta llegar al pid pasado por parametro.
*********************************************************************************************************/
void Esperar(int pid) {
}

/********************************************************************************************************
Funcin usada por Invocar() para redireccionar las operaciones de E/S
*********************************************************************************************************/
void Redireccionar(int fuente_fd, char *fuente, int destino_fd, char *destino, boolean anadir, boolean bckgrnd) {
int flags;
int fd;

  /* si se ejecuta en background y no se tiene la entrada redireccionada , por precaucion se lleva a un
     fichero que devuelve un EOF cuando se lee de l */
  if ((fuente_fd == 0) && (bckgrnd)) {
    strcpy(fuente, "/dev/null");
    fuente_fd = NIL;
  }

  if (fuente_fd != 0) {
    close(0);			/* entrada estndar */
    if (fuente_fd > 0)
      dup(fuente_fd);
    else
      if ((open(fuente, O_RDONLY, 0)) == -1) {
        fprintf(stderr,"No se puede abrir %s", fuente);
        exit(0);
      }
  }

  if (destino_fd != 1) {
    close(1);			/* salida estndar */
    if (destino_fd > 1)
      dup(destino_fd);
    else {
      flags = O_WRONLY | O_CREAT;
      if (!anadir)
        flags = O_TRUNC;
      if ((open(destino, flags, 0666)) == -1) {
        fprintf(stderr,"No se puede crear %s", destino);
        exit(0);
      }
      if (anadir)
        lseek(1, 0L, SEEK_END);
    }
  }

  /* cerramos todos los ficheros */
  for (fd = 3; fd < MAXFILES; fd++)
    close(fd);


}

/********************************************************************************************************
Funcin que invoca un comando especifico con sus argumentos.  Tambin se encarga de redirigir la entrada
o la salida si asi se indica, aadir la informacin a un fichero (modo append) o ejecutar en background.
Devuelve el pid del proceso.
*********************************************************************************************************/
int Invocar(int n_arg, char *argumentos[], int fuente_fd, char *fuente, int destino_fd, char *destino, boolean anadir, boolean bckgrnd) {
int pid;

  if (n_arg == 0)
    return 0;

  switch (pid = fork()) {
    case -1: perror("No se puede crear un nuevo hijo.");
    case 0: /* hijo */
/*            if (!bckgrnd);
              Activar_sig(); */
            Redireccionar(fuente_fd, fuente, destino_fd, destino, anadir, bckgrnd);
            fprintf(stderr,"ejecutando %s\n", argumentos[0]);
            execvp(argumentos[0], argumentos);
            fprintf(stderr,"No se ha podido ejecutar %s\n", argumentos[0]);
            exit(0);
    default: /* padre */
             wait(NULL);	/* espera a que termine el hijo */
             if ((fuente_fd > 0) && (close(fuente_fd) == -1))
               perror("Cerrar fuente");
             if ((destino_fd > 1) && (close(destino_fd)) == -1)
               perror("Cerrar destino");
             if (bckgrnd)
               printf("%d\n", pid);
             return (pid);
  }
}

/********************************************************************************************************
Funcion que procesa un comando. El comando puede acabar con '|', '&', ';' o con una nueva lnea.
Los argumentos sern pasados a un vector para su posterior llamada con execvp.
La entrada del comando puede ser:
- stdin (por defecto) (fd = 0)
- un fichero (si est presente <)
- una tubera (si el comando es precedido por |)
La salida del comando puede ser:
- stdout (por defecto) (fd = 1)
- un fichero para ser creado o truncado (>)
- un fichero para ser creado a aadirle informacin (>>)
- una tubera (si el comando es seguido de |)

*w_pid: proceso por el que hay que esperar a que termine
tuberia: indica si se necesita o no una tuberia
*pipefd: fd de la tubera

La funcin devuelve el ltimo token, porque si fuese & afectara a toda la lnea.
********************************************************************************************************/
int Comando(int *w_pid, boolean tuberia, int *pipefd){
int token;			/* token actual */
int n_arg=0;			/* n de argumentos para un comando */
int fuente_fd=0;		/* fd de la entrada */
int destino_fd=1;		/* fd de la salida */
int tuberiafd[2];		/* fd para la tuberia */
char fuente[MAXCHAR];		/* nombre del fichero fuente */
char destino[MAXCHAR];		/* nombre del fichero destino */
char *argumentos[MAXARG];	/* vector con los argumentos para un determinado comando */
boolean anadir=FALSE;			/* indica si se abre el fichero en modo append */
int pid;
int term;

  while(1) {
    switch(token = yylex()) {
    case WORD: if (n_arg == MAXARG) {
                 perror("Demasiados argumentos.");
                 break;
               }
               /* memoria para el argumento */
               if ((argumentos[n_arg] = malloc(yyleng + 1)) == NULL) {
                 perror("No hay memoria para los argumentos.");
                 break;
               }
               strcpy(argumentos[n_arg], yytext);
               n_arg++;
               continue;

    case MENOR: if (tuberia) {		/* comprueba que no ocurren cosas como <| */
                  perror("Extra <");
                  break;
                }
                /* comprueba que la entrada es un fichero */
                if (yylex() != WORD) {
                  perror("Fichero esperado");
                  break;
                }
                fuente_fd = NIL;	/* la entrada es un fichero con su fd */
                strcpy(fuente, yytext);
                continue;

    case MAYOR: if (yylex() != WORD) {
                  perror("Fichero para salida esperado.");
                  break;
                }
                destino_fd = NIL;	/* la salida es un fichero con su propio fd */
                strcpy(destino, yytext);
                break;
    case MAYOR2: if (destino_fd != 1) {
                   perror("Extra > o >>");
                   break;
                 }
                 /* comprobamos que la salida es un fichero */
                 if (yylex() != WORD) {
                   perror("Fichero para salida esperado.");
                   break;
                 }
                 destino_fd = NIL;	/* la salida es un fichero con su propio fd */
                 strcpy(destino, yytext);
                 anadir = TRUE;
                 continue;

    case BARRA: break;
    case AND: printf("\nbackground\n");
              break;
    case PYC: break;
    case CHD: yylex();	/* leemos el directorio */
              if (chdir(yytext) == -1)
                perror("chdir");
	      break;
    case LIMITE: fprintf(stderr, "\nlimite\n");
                 break;
    case NL: argumentos[n_arg] = NULL;		/* no hay mas argumentos */
             if (token == BARRA) {
               if (destino_fd != 1) {
                 perror("conflicto entre > o >> con |");
                 break;
               }
               term = Comando(w_pid, TRUE, &destino_fd);
             }
             else
               term = token;
             /* si se necesita una tuberia se crea */
             if (tuberia) {
               if (pipe(tuberiafd) == -1)
                 perror("pipe");
               *pipefd = tuberiafd[1];
               fuente_fd = tuberiafd[0];
             }

             pid = Invocar(n_arg, argumentos, fuente_fd, fuente, destino_fd, destino, anadir, (term == AND));

             if (token != BARRA)
               *w_pid = pid;
             if ((n_arg == 0) && ((token != NL) || (fuente_fd > 1)))
               perror("Comando perdido.");
             while (--n_arg >= 0)
               free(argumentos[n_arg]);
             return term;
    default: break;
    } /* fin switch */
  } /* fin while */
}


/*******************************************************************************************************/
int main(int argc, char *argv[]) {
int pid;
int fd;
int term;

 /* Ignorar_sig(); */
  signal(SIGINT, Manejador);
  printf("\n%s", PROMPT);

  while (1) {
    term = Comando(&pid,FALSE, NULL);
    /* si no es en backgraund se espera a que vayan terminando los hijos */
    if ((term != AND) && (pid != 0));
      Esperar(pid);
    if (term == NL)
      printf("%s", PROMPT);
    for (fd = 3; fd < MAXFILES; fd++)
      close(fd);
  }

  return 0;
}