#include "analizador.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#define MAXFILES 20	/* n maximo de archivos */
#define MAXCHAR 50	/* longitud mxima de nombre de fichero */
#define MAXARG 20	/* n mximo de argumentos */
#define NIL -2		/* para fd que no se usan en ese momento*/		

#define PROMPT "[OscarSH]$ "

/******************************************** TIPOS *********************************/
typedef enum {FALSE = 0, TRUE = 1} boolean;

/******************************************** VARIABLES *********************************/
extern char *yytext;	/* parte de la cadena de entrada correspondiente a un token */
extern int yyleng;
int pid_bck;
boolean old_bck = 0;

/****************************************** PROTOTIPOS *********************************/
void Pausa(void);
void Manejador();
void Ignorar();
void Esperar(int pid);
void Mostrar(int flag);
void Redireccionar(int fuente_fd, char *fuente, int destino_fd, char *destino, int error_fd, char *error,
	           boolean anadir, boolean bckgrnd);
int Llamar(int n_arg, char *argumentos[], int fuente_fd, char *fuente, int destino_fd, char *destino,
	    int error_fd, char *error, boolean anadir, boolean bckgrnd);
int Analizar(int *pid, boolean tuberia, int pipefd[2]);

/************************************************************************
Funcion que realiza una pausa en espera de que se pulse ENTER.
************************************************************************/
void Pausa(void) {

  printf("Presione ENTER para continuar...");
  getchar();
}

/********************************************************************************************************
Funcin que hace que se ignoren las seales SIGINT y SIGQUIT.
*********************************************************************************************************/
void Manejador() {
static boolean primera = TRUE;

  if (primera) {
    signal(SIGINT,Manejador);
    primera = FALSE;
    printf("\nSeal de fin recibida.Matando a todos los hijos.Pulse Ctrl+C para terminar.\n");
    if (setpgrp() == -1)
      perror("setpgrp");
    kill(0, SIGTERM);
  }
  else
    exit(0);
}

void Ignorar() {
  signal(SIGUSR1, Ignorar);
}

/********************************************************************************************************
Funcin que espera la finalizacin de procesos hasta llegar al pid pasado por parametro.
*********************************************************************************************************/
void Esperar(int pid) {
int wpid;

  while ((wpid = wait(NULL)) != pid);

}

/********************************************************************************************************
Funcin que los limites actuales de un recurso especificado en flag.
*********************************************************************************************************/
void Mostrar(int flag) {
struct rlimit *rlim;

  rlim = (struct rlimit *) malloc(sizeof(struct rlimit));

  /* obtenemos el limite */
  if (getrlimit(flag, rlim) == -1)
    perror("getrlimit");

  switch (flag) {
    case RLIMIT_CPU: printf("\nTiempo de cpu: \n\tactual:\t%d\n\tmx:\t%d\n", rlim->rlim_cur, rlim->rlim_max);
                     break;
    case RLIMIT_FSIZE: printf("\nTamao max. de fichero: \n\tactual:\t%d\n\tmx:\t%d\n", rlim->rlim_cur, rlim->rlim_max);
                       break;
    case RLIMIT_DATA: printf("\nTam. max. de la regin de datos: \n\tactual:\t%d\n\tmx:\t%d\n", rlim->rlim_cur, rlim->rlim_max);
                      break;
    case RLIMIT_STACK: printf("\nTam. max. de la region de pila: \n\tactual:\t%d\n\tmx:\t%d\n", rlim->rlim_cur, rlim->rlim_max);
                       break;
    case RLIMIT_NOFILE: printf("\nNmero max. de fd: \n\tactual:\t%d\n\tmx:\t%d\n", rlim->rlim_cur, rlim->rlim_max);
                        break;
    default: break;
  }


  free(rlim);
}

/********************************************************************************************************
Funcin usada por Llamar() para redireccionar las operaciones de E/S
*********************************************************************************************************/
void Redireccionar(int fuente_fd, char *fuente, int destino_fd, char *destino, int error_fd, char *error,
                   boolean anadir, boolean bckgrnd) {
int flags;
int fd;

  /* si se ejecuta en background y no se tiene la entrada redireccionada , por precaucion se lleva a un
     fichero que devuelve un EOF cuando se lee de l */
  if ((fuente_fd == 0) && (bckgrnd)) {
    strcpy(fuente, "/dev/null");
    fuente_fd = NIL;
  }

  if (fuente_fd != 0) {
    close(0);			/* entrada estndar */
    if (fuente_fd > 0)		/* es una tubera */
      dup(fuente_fd);
    else
      if ((open(fuente, O_RDONLY, 0)) == -1) {
        fprintf(stderr,"No se puede abrir %s", fuente);
        exit(0);
      }
  }

  if (destino_fd != 1) {
    close(1);			/* salida estndar */
    if (destino_fd > 1)		/* salida a una tubera */
      dup(destino_fd);
    else {			/* salida a un fichero */
      if (anadir) {
        if ((open(destino,O_WRONLY)) == -1) {
          fprintf(stderr,"No se puede crear %s", destino);
          exit(0);
        }
        lseek(1, 0L, SEEK_END);
      }
      else
        if ((creat(destino,S_IREAD|S_IWRITE)) == -1) {
          fprintf(stderr,"No se puede crear %s", destino);
          exit(0);
        }
    }
  }

  if (error_fd != 2) {
    close(2);			/* salida error estndar */
    if (error_fd > 2)
      dup(error_fd);
    else if ((creat(error,S_IREAD|S_IWRITE)) == -1) {
          fprintf(stderr,"No se puede crear %s para errores", error);
          exit(0);
    }
  }

  /* cerramos todos los ficheros que no vamos a necesitar */
  for (fd = 3; fd < MAXFILES; fd++)
    close(fd);
}

/********************************************************************************************************
Funcin que invoca un comando especifico con sus argumentos.  Tambin se encarga de redirigir la entrada
o la salida si asi se indica, aadir la informacin a un fichero (modo append) o ejecutar en background.
Devuelve el pid del proceso.
*********************************************************************************************************/
int Llamar(int n_arg, char *argumentos[], int fuente_fd, char *fuente, int destino_fd, char *destino,
	    int error_fd, char *error, boolean anadir, boolean bckgrnd) {
int pid;

  if (n_arg == 0)
    return 0;

  switch (pid = fork()) {
    case -1: perror("No se puede crear un nuevo hijo.");
    case 0: /* hijo */
            Redireccionar(fuente_fd, fuente, destino_fd, destino, error_fd, error, anadir, bckgrnd);
/*            fprintf(stderr,"ejecutando %s\n", argumentos[0]); */
            execvp(argumentos[0], argumentos);
            fprintf(stderr,"%s: ", argumentos[0]);
            perror("");
            exit(0);
    default: /* padre */
             if (bckgrnd) {
               printf("%d\n", pid);
               old_bck = TRUE;
               pid_bck = pid;
             }
/*             else if ((fuente_fd == 0) || (destino_fd == 1))
               wait(NULL); */	/* espera a que termine el hijo */

             if ((fuente_fd > 0) && (close(fuente_fd) == -1))
               perror("Cerrar fuente");
             if ((destino_fd > 1) && (close(destino_fd)) == -1)
               perror("Cerrar destino");

             return (pid);
  }
}

/********************************************************************************************************
Funcion que procesa un comando. El comando puede acabar con '|', '&', ';' o con una nueva lnea.
Los argumentos sern pasados a un vector para su posterior llamada con execvp.
La entrada del comando puede ser:
- stdin (por defecto) (fd = 0)
- un fichero (si est presente <)
- una tubera (si el comando es precedido por |)
La salida del comando puede ser:
- stdout (por defecto) (fd = 1)
- un fichero para ser creado o truncado (>)
- un fichero para ser creado a aadirle informacin (>>)
- una tubera (si el comando es seguido de |)

*w_pid: proceso por el que hay que esperar a que termine
tuberia: indica si se necesita o no una tuberia
*pipefd: fd de la tubera

La funcin devuelve el ltimo token, porque si fuese & afectara a toda la lnea.
********************************************************************************************************/
int Analizar(int *w_pid, boolean tuberia, int pipefd[2]){
int token;			/* token actual */
int n_arg=0;			/* n de argumentos para un comando */
int fuente_fd=0;		/* fd de la entrada */
int destino_fd=1;		/* fd de la salida */
int error_fd = 2;		/* fd de la salida de errores */
int tuberiafd[2];		/* fd para la tuberia */
char fuente[MAXCHAR];		/* nombre del fichero fuente */
char destino[MAXCHAR];		/* nombre del fichero destino */
char error[MAXCHAR];		/* nombre del fichero de errores */
char *argumentos[MAXARG];	/* vector con los argumentos para un determinado comando */
boolean anadir=FALSE;		/* indica si se abre el fichero en modo append */
int pid;
int term;
boolean bckgrnd = FALSE;	/* indica si es en background o no */
struct rlimit rlim;

  while(1) {
    switch(token = yylex()) {

    case WORD: if (n_arg == MAXARG) {
                 perror("Demasiados argumentos.");
                 break;
               }
               /* memoria para el argumento */
               if ((argumentos[n_arg] = malloc(yyleng + 1)) == NULL) {
                 perror("No hay memoria para los argumentos.");
                 break;
               }
               strcpy(argumentos[n_arg], yytext);
               n_arg++;
               break;			/* continue; */

    case MENOR: if (tuberia) {		/* comprueba que no ocurren cosas como <| */
                  perror("Extra <");
                  break;
                }
                /* comprueba que la entrada es un fichero */
                if (yylex() != WORD) {
                  perror("Fichero esperado");
                  break;
                }
                fuente_fd = NIL;	/* la entrada es un fichero con su fd */
                strcpy(fuente, yytext);
                break;			/*continue; */

    case MAYOR: if (yylex() != WORD) {
                  perror("Fichero para salida esperado.");
                  break;
                }
                destino_fd = NIL;	/* la salida es un fichero con su propio fd */
                strcpy(destino, yytext);
                break;

    case MAYOR2: if (destino_fd != 1) {
                   perror("Extra > o >>");
                   break;
                 }
                 /* comprobamos que la salida es un fichero */
                 if (yylex() != WORD) {
                   perror("Fichero para salida esperado.");
                   break;
                 }
                 destino_fd = NIL;	/* la salida es un fichero con su propio fd */
                 strcpy(destino, yytext);
                 anadir = TRUE;
                 break; 		/* continue; */

    case MAYORAND: if (yylex() != WORD) {
                     perror("Fichero para errores esperado.");
                     break;
                   }
                   error_fd = NIL;	/* la salida es un fichero con su propio fd */
                   strcpy(error, yytext);
                   break;

    case BARRA: if (destino_fd != 1) {
                   perror("conflicto entre > o >> con |");
                   break;
                }
                argumentos[n_arg] = NULL;		/* no hay mas argumentos */
                if (pipe(tuberiafd) == -1)
                    perror("pipe");
                /* si viene de una tuberia ya creada */
                if (tuberia) {
                  fuente_fd = pipefd[0];
                  destino_fd = tuberiafd[1];
                }
                else		/* sin no estaba creada */
                  destino_fd = tuberiafd[1];  		/* escritura */

/*               fuente_fd = tuberiafd[0];*/		/* lectura */
                term = Analizar(w_pid, TRUE, tuberiafd);
                pid = Llamar(n_arg, argumentos, fuente_fd, fuente, destino_fd, destino, error_fd, error,
			    anadir, bckgrnd);
                tuberia = TRUE;

                return term;
    case AND: bckgrnd = TRUE;
              break;
    case PYC: break;
/*
              pid = Llamar(n_arg, argumentos, fuente_fd, fuente, destino_fd, destino, error_fd, error,
			    anadir, bckgrnd);
              waitpid(pid);
              term = Analizar(w_pid, TRUE, tuberiafd);
              while (--n_arg >= 0)
               free(argumentos[n_arg]);
              break;*/
    case CHD: yylex();	/* leemos el directorio */
              if (chdir(yytext) == -1)
                perror("chdir");
	      break;
    case LIMITE: switch(token = yylex()) {
                   case CPU: if ((token = yylex()) != WORD)	/* si no viene el limite */
                               /* se muestra el valor actual de CPU */
                               Mostrar(RLIMIT_CPU);
                             else {   				/* se establece el valor mximo */
                               rlim.rlim_max = atoi(yytext);
                               setrlimit(RLIMIT_CPU, &rlim);

                             }
                             break;
                   case FSIZE: if ((token = yylex()) != WORD)	/* si no viene el limite */
                                 /* se muestra el valor actual de FSIZE */
                                 Mostrar(RLIMIT_FSIZE);
                               else {   				/* se establece el valor mximo */
                                 rlim.rlim_max = atoi(yytext);
                                 setrlimit(RLIMIT_FSIZE, &rlim);
                               }
                               break;
                   case DATA: if ((token = yylex()) != WORD)	/* si no viene el limite */
                                /* se muestra el valor actual de DATA */
                                Mostrar(RLIMIT_DATA);
                              else {   				/* se establece el valor mximo */
                                rlim.rlim_max = atoi(yytext);
                                setrlimit(RLIMIT_DATA, &rlim);
                              }
                              break;
                   case STACK: if ((token = yylex()) != WORD)	/* si no viene el limite */
                                 /* se muestra el valor actual de STACK */
                                 Mostrar(RLIMIT_STACK);
                               else {   				/* se establece el valor mximo */
                                 rlim.rlim_max = atoi(yytext);
                                 setrlimit(RLIMIT_STACK, &rlim);
                               }
                               break;
                   case NUMFICH: if ((token = yylex()) != WORD)	/* si no viene el limite */
                                  /* se muestra el valor actual de NUMFICH */
                                  Mostrar(RLIMIT_NOFILE);
                                else {   				/* se establece el valor mximo */
                                  rlim.rlim_max = atoi(yytext);
                                  setrlimit(RLIMIT_NOFILE, &rlim);
                                }
                                break;
                   default: /* si no se especifica ningn recurso se presentan todos los lmites */
                            Mostrar(RLIMIT_CPU);
                            Mostrar(RLIMIT_FSIZE);
                            Mostrar(RLIMIT_DATA);
                            Mostrar(RLIMIT_STACK);
                            Mostrar(RLIMIT_NOFILE);
                            break;
                 }
                 return LIMITE;
    case NL: /*if (old_bck) 			/* si hay un proceso ejecutandose en background */
/*               if (kill(pid_bck, SIGUSR1) == -1) {		/* ya ha acabado */
/*                 printf("\n\tTerminado %d\n\n", pid_bck);
                 old_bck = FALSE;
               } */

             argumentos[n_arg] = NULL;		/* no hay mas argumentos */

             if (tuberia)	/* si viene de una tuberia ya creada */
               fuente_fd = pipefd[0];

/*             printf("Bck %d\n", bckgrnd);*/
             pid = Llamar(n_arg, argumentos, fuente_fd, fuente, destino_fd, destino, error_fd, error,
	 	           anadir, bckgrnd);

             *w_pid = pid;

             if ((n_arg == 0) && ((token != NL) || (fuente_fd > 1)))
               perror("Comando perdido.");

             while (--n_arg >= 0)
               free(argumentos[n_arg]);

             if (bckgrnd)
               return AND;

             return token;
    default: break;
    } /* fin switch */
  } /* fin while */
}


/*******************************************************************************************************/
int main(int argc, char *argv[]) {
int pid;
int fd;
int term;
int pipefd[2];

 /* Ignorar_sig(); */
  signal(SIGINT, Manejador);
  signal(SIGUSR1, Ignorar);
  printf("\n%s", PROMPT);

  while (1) {
    term = Analizar(&pid,FALSE, pipefd);
    /* si no es en backgraund se espera a que vayan terminando los hijos */
    if ((term != AND) && (pid != 0) && (term != LIMITE))
      Esperar(pid);

/*    if ((term == NL) || (term == LIMITE)) */
      printf("%s", PROMPT);
    for (fd = 3; fd < MAXFILES; fd++)
      close(fd);
  }

  return 0;
}
