/*******************************************************************************************************
 Oscar Martnez Valpuesta - 4 Ingeniera Superior Informtica
 C.A.S.O. Practica 3 - Aplicacin cliente servidor usando sockets de UNIX
 Descripcin: 	Se pretende realizar una version de remote shell, una aplicacin que permita ejecutar
		comandos remotamente. Para ello se crear un programa que har las funciones de servidor
		y que ser el encargado de ejecutar el comando que le enva desde una mquina remota
		otro programa cliente.
			Este es el mdulo del servidor. Los pasos a seguir son:
		* creacin del socket --> socket
		* asociacion del servicio con el socket --> bind
		* escucha del puerto por si alguien se conecta --> listen
		* aceptacion de la conexion -->	accept
		* intercambio de datos: la entrada se analiza con la herramienta flex. El cliente enva
			una cadena que se descompone en los diferentes argumentos con la llamada a la
			funcin yylex() de flex.
*******************************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>
#include "analizador.h"

#define PUERTO 5000
#define PUERTO_CSI 6571

#define LIMITE 5
#define MAXDATOS 100
#define MAXARG 20		/* numero maximo de argumentos */


extern char *yytext;
extern int yyleng;
extern char *myinputptr;		/* posicion en la cadena de entrada */
extern char *myinputlim;		/* tamao de la entrada */

int n_sockfd;			/* socket despues de accept */


void Manejador() {

  fprintf(stderr,"Cerrando conexin\n");
  exit(1);
}

/*******************************************************************************************************
Funcion que crea el socket, lo asocia al servicio y se pepara para escuchar.
Devuelve el socket o -1 si ha habido algun error.
*******************************************************************************************************/
int Servicio(void) {
struct sockaddr_in mi_addr;	/* informacin sobre mi direccin */
int sockfd;			/* socket file descriptor */
struct servent *servicio;	/* estructura de un servicio */

  /* obtenemos el puerto para un servicio caso/TCP (para que sea portable el codigo */
  servicio = (struct servent *)getservbyname("caso", "tcp");
  if (servicio == NULL) {
    perror("ERROR en getservbyname");
    return(-1);
  }

  /* creamos el socket */
  if ((sockfd = socket(AF_INET, SOCK_STREAM,0)) == -1) {
    perror("creacion socket");
    return -1;
  }

  /* asociamos el socket al servicio */
  mi_addr.sin_family = AF_INET;
/*  mi_addr.sin_port = htons(PUERTO_CSI);  */
  mi_addr.sin_port = htons(PUERTO);
  /* puerto para el servicio caso */
/*  mi_addr.sin_port = servicio->s_port; */
/*  mi_addr.sin_addr.s_addr = gethostbyname("gofio.csi.ull.es"); */
  mi_addr.sin_addr.s_addr = INADDR_ANY;
  bzero(&(mi_addr.sin_zero),8);

  if ((bind(sockfd, (struct sockaddr *)&mi_addr, sizeof(struct sockaddr))) == -1) {
    perror("bind");
    return -1;
  }

  /* nos ponemos a escuchar */
  if ((listen(sockfd, LIMITE)) == -1) {
    perror("listen");
    return -1;
  }

  return sockfd;
}

/*******************************************************************************************************/
int main(int argc, char *argv) {
int sockfd;
struct sockaddr_in su_addr;	/* informacin de la mquina que se conecta */
int tamano;			/* tamao de la estructura */
char buffer[MAXDATOS];		/* cadena recibida desde el cliente */
int pid;			/* pid despues de hacer fork() */
char *argumentos[MAXARG];	/* argumentos en la entrada */
int n_arg;			/* numero de argumentos */
int token;			/* token actual */
int fin;			/* indica final de la entrada */


  signal (SIGINT, Manejador);

  if ((sockfd = Servicio()) == -1)
    exit(1);

  /* bucle accept */
  while (1) {
/*    printf("Esperando recibir conexin...\n"); */
    tamano = sizeof(struct sockaddr_in);
    if ((n_sockfd = accept(sockfd, (struct sockaddr *)&su_addr, &tamano)) == -1)
      perror("accept");
    printf("Servidor: conexin desde %s\n", inet_ntoa(su_addr.sin_addr));

    pid = fork();
    if (pid == 0) {
      /* aqui no se puede cerrar la salida estandar y duplicar prque todavia hay que leer del
         socket y daria problemas porque lo estariamos haciendo desde stdout */
      while(1) {
/*          fprintf(stderr,"Esperando recibir datos...\n"); */
          tamano = read(n_sockfd, buffer, MAXDATOS);
          if (strcmp(buffer,"exit") == 0)
            exit(10);
          else { 				/* analizamos la entrada */
            n_arg = 0;
            fin = 0;
            /* hacemos que lex vaya tomando la entrada de buffer*/
            myinputptr = buffer;
            myinputlim = myinputptr + strlen(buffer);
            while (!fin)
            switch(token = yylex()) {
              case WORD: if (n_arg == MAXARG) {
                           perror("Demasiados argumentos.");
                           break;
                         }
                         /* memoria para el argumento */
                         if ((argumentos[n_arg] = malloc(yyleng + 1)) == NULL) {
                           perror("No hay memoria para los argumentos.");
                            break;
                         }
                         strcpy(argumentos[n_arg], yytext);
                         n_arg++;
                         continue;
              case 0: fin = 1;
                       break;
              case EXIT: exit(1);
              default: break;
            } /* end switch */
          }
          argumentos[n_arg] = NULL;
        /* creamos un hijo para la ejecucion de un comando */
        pid = fork();
        if (pid == 0) {
          close(1);			/* salida estandar */
          dup(n_sockfd);
          close(n_sockfd);
          execvp(argumentos[0], argumentos);
          /* si no se ha podido ejecutar el comando se manda el error al cliente */
          write(1,strerror(errno), strlen(strerror(errno)));
          exit(1);
        }
        else {
          if(wait(NULL) == -1)
            perror("wait");
          write(n_sockfd, "FIN", 4);
          /* liberamos la memoria de los argumentos */
          while (--n_arg >= 0)
            free(argumentos[n_arg]);
        }
      } /* while */
    } /* fork */

    close(n_sockfd);
  } /* accept */

  return 0;
}
