/*************************************************************************************
  Autor: Oscar Martnez Valpuesta -- 4 Ingenieria Superior Informtica
  Descripcin: Mdulo que contiene las llamadas de alto nivel.
		* abrir --> open
		* cerrar --> close
		* leer --> read
		* escribir --> write
		* posicionar --> lseek
*************************************************************************************/
#include "callHi.h"

/***************************************************************************************
Accion a realizar cuando se recibe la seal SIGUSR2 -> escritura retardada. Se escribe
el buffer correspondiente a disco.
****************************************************************************************/
void escritura_r() {
int i,j;
long int pos;
int tam;
int aux=0;

  signal(SIGUSR2, escritura_r);

  /* buscamos el buffer de escritura reatardada */
  for (i = 0; i < MAXBUF; i++)
    if ((buffer_cache->buf[i].retardada) && (buffer_cache->buf[i].pid = getpid())) {
      buffer_cache->buf[i].retardada = FALSE;
      pos = buffer_cache->buf[i].bloque * TAMBUF;
      lseek(buffer_cache->buf[i].fd, pos, SEEK_SET);
      tam = buffer_cache->buf[i].tamano;
      aux = buffer_cache->buf[i].inicial;
      for(j = 0; j < tam; j++)
        write(buffer_cache->buf[i].fd, &buffer_cache->buf[i].datos[(j+aux)], 1);
      buffer_cache->buf[i].validos = TRUE;
      buffer_cache->buf[i].bloqueado = FALSE;
      buffer_cache->buf[i].tamano = 0;
      buffer_cache->buf[i].inicial = NIL;
      est.retardados++;
      PutFinFreeList(i);
    }
}


/***************************************************************************************
Accion a realizar cuando se recibe la seal SIGUSR1 -> despertar
****************************************************************************************/
void despertar() {
  signal(SIGUSR1, despertar);
}

/***************************************************************************************
Funcin que inicializa la tabla de ficheros de cada proceso. Esta tabla es nica y no
compartida para cada proceso.
Tambien inicialia las estadisticas.
****************************************************************************************/
void Inicializar(void) {
int i;

  for (i = 0; i < MAXFILES; i++)
    tabla_descriptores[i].entrada = NIL;

  est.encontrados = 0;
  est.liberados = 0;
  est.retardados = 0;
  est.dormidos = 0;
  est.nuevos = 0;
}


/***************************************************************************************
Funcin que inserta en la tabla de ficheros y en la tabla de descriptores de ficheros
de usuario el fd. Las entradas marcadas con NIL estan libres.
Devuelve la entrada en la tabla de descriptores de usuario donde situar los datos.
Devuelve -1 si no hay ms espacio.
****************************************************************************************/
int InsertaTabla(int fd_real) {
int i = 0;
int j = 0;
boolean insertado = FALSE;
boolean libre = FALSE;

  /* buscamos posicion en la tabla de ficheros */
  while ((j < MAXFILES) && (!libre)) {
    if (tabla_ficheros->entrada[j].offset == NIL)
      libre = TRUE;
    j++;
  }

  if (libre) {
    j--;
    /* buscamos posicin en la tabla de descriptores */
    while ((i < MAXFILES) && (!insertado) ) {
      if (tabla_descriptores[i].entrada == NIL) {
        insertado = TRUE;
        tabla_descriptores[i].fd_real = fd_real;
        tabla_descriptores[i].entrada = j;
        return i;
      }

      i++;
    }
  }

  return -1;
}


/***************************************************************************************
Procedimiento que muestr las estadisticas en pantalla
****************************************************************************************/
void Mostrar(void) {

  printf("N de asignaciones nuevas buffer-->bloque: %ld\n", est.nuevos);
  printf("N de accesos encontrados en el buffer cach: %ld\n", est.encontrados);
  printf("N de veces dormido: %ld\n", est.dormidos);
  printf("N de escrituras retardadas: %ld\n", est.retardados);
  printf("N de buffers liberados: %ld\n", est.liberados);
}


/***************************************************************************************
Funcin que lee un n_bytes del fichero con descriptor fd a la direccin de memoria
apuntada por buffer_user.
Devuelve el numero de bytes leidos.
****************************************************************************************/
int Leer(int fd, char *buffer_user, int n_bytes) {
int contador=0;		/* numero de bytes leidos hasta el momento */
unsigned long int inodo;
int buffer;
int bytes_a_leer;
unsigned int disp;
unsigned long int  bloque;
int fd_real;
int entrada;
int i;
unsigned int inicio;		/* dentro del buffer, posicion desde la que se inicia la copia */
unsigned int fin;		/* final dentro del buffer a copiar */

  /* obtenemos el fd real y la entrada en la tabla de ficheros a partir de la */
  /* tabla de descriptores del proceso */
  fd_real = tabla_descriptores[fd].fd_real;
  entrada = tabla_descriptores[fd].entrada;

  /* chequeamos la accesibilidad de lectura */
  if (!tabla_ficheros->entrada[entrada].lectura) {	/* si no hay permisos de lectura */
    printf("ERROR, no hay permisos de lectura.\n");
    exit (1);
  }
  inodo = tabla_ficheros->entrada[entrada].inodo;
  disp = tabla_ficheros->entrada[entrada].disp;

  while (contador < n_bytes) {
    /* calculamos el bloque de disco correspondiente al offset actual */
    bloque = tabla_ficheros->entrada[entrada].offset / TAMBUF;
    /* calculamos los bytes a leer dentro del bloque actual */
    bytes_a_leer = (bloque+1) * TAMBUF - tabla_ficheros->entrada[entrada].offset;
    if (bytes_a_leer > (n_bytes-contador))
      bytes_a_leer = n_bytes - contador;
    if (bytes_a_leer > (tabla_ficheros->entrada[entrada].tam - tabla_ficheros->entrada[entrada].offset))
      bytes_a_leer = tabla_ficheros->entrada[entrada].tam - tabla_ficheros->entrada[entrada].offset;


    if ((tabla_ficheros->entrada[entrada].offset == tabla_ficheros->entrada[entrada].tam) || (bytes_a_leer == 0))
      return contador;

    buffer = bread(fd_real, disp, inodo, bloque);		/* devuelve un buffer */

    /* copiamos los datos al buffer del ususario */
    if ((tabla_ficheros->entrada[entrada].offset % TAMBUF) == 0) {
      inicio = 0;
      fin = bytes_a_leer;
    }
    else {
      inicio = tabla_ficheros->entrada[entrada].offset % TAMBUF;	/*TAMBUF - bytes_a_leer; */
      fin = inicio + bytes_a_leer;
    }

    for (i = inicio; i < fin; i++) {
      *buffer_user = buffer_cache->buf[buffer].datos[i];
      buffer_user += 1;
    }

    /* actualizamos el contador y el offset */
    contador += bytes_a_leer;
    tabla_ficheros->entrada[entrada].offset += bytes_a_leer;
    /* liberamos el buffer */
    brelse(buffer);
  }

  return contador;
}

/***************************************************************************************
Funcin que crea un fichero con nombre name con el modo de apertura flags.
flags puede valer:
0 LECTURA
1 ESCRITURA
2 L_E (lectura-escritura)
Devuelve el file descriptor del fichero.
Devuelve -1 si no hay ms espacio en la tabla de ficheros.
****************************************************************************************/
int Crear(char *name, int flags) {
struct stat inodo;    			/* estructura del inodo */
int entrada;				/* entrada de la tabla de ficheros */
boolean lectura = FALSE;
boolean escritura = FALSE;
int fd_real;				/* devuelto por creat */
int fd;					/* para nuestra tabla de descriptores */

  switch(flags) {
    case 0: lectura = TRUE;
            fd_real = creat(name, S_IREAD);
            break;
    case 1: escritura = TRUE;
            fd_real = creat(name, S_IWRITE);
            break;
    case 2: lectura = TRUE;
            escritura = TRUE;
            fd_real = creat(name, S_IREAD|S_IWRITE);
            break;
    default: printf("Modo de creacin no vlido.\n");
             exit(1);
  }

  if (fd_real == -1)
    return -1;

  if ((fd = InsertaTabla(fd_real)) == -1)
    return -1;

  /* accedemos al inodo que se le ha dado */
  if (stat(name, &inodo) == -1)
    perror("acceso inodo (crear)");

  entrada = tabla_descriptores[fd].entrada;
  tabla_ficheros->entrada[entrada].inodo = inodo.st_ino;
  tabla_ficheros->entrada[entrada].disp = inodo.st_dev;
  tabla_ficheros->entrada[entrada].offset = 0;
  tabla_ficheros->entrada[entrada].lectura = lectura;
  tabla_ficheros->entrada[entrada].escritura = escritura;

  return fd;
}

/***************************************************************************************
Funcin que abre un fichero con nombre name con el modo de apertura flags.
flags puede valer:
0 LECTURA
1 ESCRITURA
2 L_E (lectura-escritura)
Devuelve el file descriptor del fichero.
Devuelve -1 si no hay ms espacio en la tabla de ficheros.
****************************************************************************************/
int Abrir(char *name, int flags) {
struct stat inodo;    			/* estructura del inodo */
int entrada;				/* entrada de la tabla de ficheros */
boolean lectura = FALSE;
boolean escritura = FALSE;
int fd_real;				/* devuelto por open */
int fd;					/* para nuestra tabla de descriptores */

  /* obtenemos los datos del inodo */
  if (stat(name, &inodo) == -1)
    perror("acceso inodo (abrir)");

  /* si es el propietario el que los quiere abrir */
  if (inodo.st_uid == getuid()) {
    /* comprobamos los permisos del propietario */
    switch((inodo.st_mode & 0x1FF) >> 6) {
      case 4: lectura = TRUE;
              break;
      case 2: escritura = TRUE;
              break;
      case 6: lectura = TRUE;
              escritura = TRUE;
              break;
      case 5: lectura = TRUE;
              break;
      case 3: escritura = TRUE;
              break;
      case 7: lectura = TRUE;
              escritura = TRUE;
              break;
      default: printf("ERROR al obtener los permisos del inodo.\n");
               exit(1);
    }
  }
  else
    /* es alguien del grupo */
    if (inodo.st_gid == getgid()) {
      /* comprobamos los permisos de acceso grupo */
      switch((inodo.st_mode & 56) >> 3) {
        case 4: lectura = TRUE;
                break;
        case 2: escritura = TRUE;
                break;
        case 6: lectura = TRUE;
                escritura = TRUE;
                break;
        case 5: lectura = TRUE;
                break;
        case 3: escritura = TRUE;
                break;
        case 7: lectura = TRUE;
                escritura = TRUE;
                break;
        default: printf("ERROR al obtener los permisos del inodo.\n");
                 exit(1);
      }
    }
    else{ /* intenta acceder cualquiera */
      /* comprobamos los permisos de acceso de cualquiera */
      switch(inodo.st_mode & 0x007) {
        case 4: lectura = TRUE;
                break;
        case 2: escritura = TRUE;
                break;
        case 6: lectura = TRUE;
                escritura = TRUE;
                break;
        case 5: lectura = TRUE;
                break;
        case 3: escritura = TRUE;
                break;
        case 7: lectura = TRUE;
                escritura = TRUE;
                break;
        default: printf("ERROR al obtener los permisos del inodo.\n");
                 exit(1);
      }
    }

  /* comprobamos modo de apertura con permisos */
  if (((flags == LECTURA) && !(lectura)) || ((flags == ESCRITURA) && !(escritura)) || ((flags == L_E) && !(lectura) && !(escritura))) {
    printf("ERROR, no se tienen los permisos correctos.\n");
    exit(1);
  }

  if ((fd_real = open(name,flags)) == -1)
    perror("open");

  if ((fd = InsertaTabla(fd_real)) == -1)
    return -1;

  entrada = tabla_descriptores[fd].entrada;
  tabla_ficheros->entrada[entrada].inodo = inodo.st_ino;
  tabla_ficheros->entrada[entrada].disp = inodo.st_dev;
  tabla_ficheros->entrada[entrada].tam = inodo.st_size;
  tabla_ficheros->entrada[entrada].offset = 0;
  tabla_ficheros->entrada[entrada].lectura = lectura;
  tabla_ficheros->entrada[entrada].escritura = escritura;

  return fd;
}

/***************************************************************************************
Funcin que cierra el fichero con el descriptor fd. fd es el descriptor real y hay que
liberar (poner a NIL) tanto en la tabla de ficheros del proceso como en la tabla de ficheros real.
Devuelve -1 si ha ocurrido algn error.
****************************************************************************************/
int Cerrar(int fd){
int i;
int fd_real;
int entrada;
int aux;

  /* obtenemos el fd real y la entrada en la tabla de ficheros a partir de la */
  /* tabla de descriptores del proceso */
  fd_real = tabla_descriptores[fd].fd_real;
  entrada = tabla_descriptores[fd].entrada;

  if (fd_real == NIL)
    return -1;

  /* miramos si queda algn buffer por escribir */
  for (i = 0; i < MAXBUF; i++)
    if (buffer_cache->buf[i].retardada)
      if ((kill(buffer_cache->buf[i].pid, SIGUSR2)) == -1)
	  perror("seal escritura retardada");

  /* liberamos en la tabla de ficheros del proceso */
  tabla_descriptores[fd].entrada = NIL;

  /* liberamos en la tabla de ficheros */
  tabla_ficheros->entrada[entrada].offset = NIL;

  return (close(fd_real));
}

/***************************************************************************************
Funcin que escribe un numero de bytes n_bytes al fichero con descriptor fd desde la
direccin de memoria apuntada por buffer_user.
Devuelve el numero de bytes leidos.
****************************************************************************************/
int Escribir(int fd, char *buffer_user, int n_bytes){
int contador=0;		/* numero de bytes leidos hasta el momento */
unsigned long int inodo;
int buffer;
int bytes_a_escribir;
unsigned int disp;
unsigned long int bloque;
int fd_real;
int entrada;
unsigned int inicio;		/* dentro del buffer, posicion desde la que se inicia la copia */
unsigned int fin;		/* final dentro del buffer a copiar */
int i;

  /* obtenemos el fd real y la entrada en la tabla de ficheros a partir de la */
  /* tabla de descriptores del proceso */
  fd_real = tabla_descriptores[fd].fd_real;
  entrada = tabla_descriptores[fd].entrada;

  /* chequeamos la accesibilidad de escritura */
  if (!tabla_ficheros->entrada[entrada].escritura) {	/* si no hay permisos de escritura */
    printf("ERROR, no hay permisos de escritura.\n");
    exit (1);
  }
  inodo = tabla_ficheros->entrada[entrada].inodo;
  disp = tabla_ficheros->entrada[entrada].disp;

  while (contador < n_bytes) {
    /* calculamos el bloque de disco correspondiente al offset actual */
    bloque = tabla_ficheros->entrada[entrada].offset / TAMBUF;
    /* calculamos los bytes a escribir dentro del bloque actual */
    bytes_a_escribir = (bloque+1) * TAMBUF - tabla_ficheros->entrada[entrada].offset;
    if (bytes_a_escribir > (n_bytes-contador))
      bytes_a_escribir = n_bytes - contador;
    if (bytes_a_escribir == 0)
      break;

    if (bytes_a_escribir == TAMBUF) {		/* escritura completa */
      buffer = getblk(disp,inodo,bloque);
      inicio = 0;
      fin = bytes_a_escribir;
    }
    else {				        /* escritura parcial */
      buffer = bread(fd_real, disp, inodo, bloque);		/* devuelve un buffer */
      inicio = tabla_ficheros->entrada[entrada].offset % TAMBUF;
      fin = inicio + bytes_a_escribir;
    }

    /* copiamos los datos desde el buffer del ususario */
    for (i = inicio; i < fin; i++) {
      buffer_cache->buf[buffer].datos[i]= *buffer_user ;
      buffer_user += 1;
    }

    /* datos buffer */
    buffer_cache->buf[buffer].retardada = TRUE;
    buffer_cache->buf[buffer].validos = TRUE;
    buffer_cache->buf[buffer].inodo = inodo;
    buffer_cache->buf[buffer].disp = disp;
    if (buffer_cache->buf[buffer].tamano < TAMBUF) {
      buffer_cache->buf[buffer].tamano += bytes_a_escribir;
      if (buffer_cache->buf[buffer].inicial == NIL)
        buffer_cache->buf[buffer].inicial = inicio;
    }
    buffer_cache->buf[buffer].fd = fd_real;
    buffer_cache->buf[buffer].pid = getpid();

    /* actualizamos el contador y el offset */
    contador += bytes_a_escribir;
    tabla_ficheros->entrada[entrada].offset += bytes_a_escribir;
    /* liberamos el buffer */
    brelse(buffer);
  }

  return contador;
}
