/*************************************************************************************
  Autor: Oscar Martnez Valpuesta -- 4 Ingenieria Superior Informtica
  Descripcin:

		El acceso a la cola hash se realiza mediante el n de inodo mod MAXHASH
*************************************************************************************/
#include "callLo.h"
int semid;	/* identificador del semaforo */
/*********************************************************************
Funcion que opera el semaforo sem segun la variable abrir:
TRUE -> abrir semaforo
FALSE -> cerrar semaforo
Sem puede tomar los valores 0 y 1. 0 indica el semaforo del getblk
y 1 indica el semaforo del brelse.
*********************************************************************/
void Semaforo(boolean abrir){
struct sembuf operacion;

  operacion.sem_num = 0;
  operacion.sem_flg = 0;
  if (abrir)
    operacion.sem_op = 1;
  else
    operacion.sem_op = -1;

 while((semop(semid, &operacion, 1)) == -1);
}


/*********************************************************************
Funcion que da memoria de la zona de memoria compartida al buffer
cache. Devuelve -1 si ha ocurrido algun error.
Tambien crea el semaforo
*********************************************************************/
int Da_Memoria(void) {
int id;		/* identificador de la memoria compartida del buffer cache*/
int id2;	/* identificador de la memoria compartida para la tabla de ficheros*/

  /* creamos la zona de memoria compartida. Se crea con permisos de lectura
  y modificacion para el usuario */
  id = shmget(LLAVE, TAM, IPC_CREAT | 0600);
  if (id == (-1))
    return(-1);

  id2 = shmget(LLAVE2, TAM, IPC_CREAT | 0600);
  if (id2 == (-1))
    return(-1);

  if ((semid = semget(LLAVE3,1,IPC_CREAT|0600)) == -1)
    perror("semaforo");

  buffer_cache = (BUFFER_CACHE *)shmat(id, 0,0);
  tabla_ficheros = (TABLA_FICHEROS *)shmat(id2, 0,0);

  return 0;
}



/**************************************************************************
Funcin que inserta en el vector de dormidos de un buffer especfico el
pid de un proceso a dormir. Lo inserta en la primera posicin que est
libre (marcada con NIL). Si no puede hacerlo devuelve -1
**************************************************************************/
int Insertar(unsigned long int *vector, unsigned long int pid){
int i;

  for (i = 0;i < MAXDORM; i++)
    if (vector[i] == NIL) {
      vector[i] = pid;
      return 1;
    }

  return (-1);
}

/*********************************************************************
Funcin que devuelve el primer buffer de la lista libre.
**********************************************************************/
int PopFreeList(void) {
int buffer;

  buffer = buffer_cache->freeInit;
  buffer_cache->freeInit = buffer_cache->buf[buffer].sigFree;
  /* si era el nico */
  if (buffer_cache->freeInit == NIL)
    buffer_cache->freeEnd = NIL;
  else
    buffer_cache->buf[buffer_cache->freeInit].antFree = NIL;

  return buffer;
}

/*********************************************************************
Funcin que inserta el buffer al principio de la free list.
**********************************************************************/
void PutInicioFreeList(int buffer) {
int aux;
int aux2;

  /* miramos si no est ya en la lista libre */
  aux = buffer_cache->freeInit;
  while ((aux != NIL) && (aux != buffer)) {
    aux2 = buffer_cache->buf[aux].sigFree;
    aux = aux2;
  }

  if (aux == NIL) {

    buffer_cache->buf[buffer].antFree = NIL;
    buffer_cache->buf[buffer].sigFree = buffer_cache->freeInit;

    if (buffer_cache->freeInit != NIL) {		/* lista libre no esta vacia */
      aux = buffer_cache->freeInit;
      buffer_cache->buf[aux].antFree = buffer;
    }
    else
      buffer_cache->freeEnd = buffer;

    buffer_cache->freeInit = buffer;
  }
}

/*********************************************************************
Funcin que inserta el buffer al final de la free list.
**********************************************************************/
void PutFinFreeList(int buffer){
int aux;
int aux2;

  /* miramos si no est ya en la lista libre */
  aux = buffer_cache->freeInit;

  while ((aux != NIL) && (aux != buffer)) {
    aux2 = buffer_cache->buf[aux].sigFree;
    aux = aux2;
  }


  if (aux == NIL) {
    buffer_cache->buf[buffer].sigFree = NIL;
    buffer_cache->buf[buffer].antFree = buffer_cache->freeEnd;

    if (buffer_cache->freeEnd != NIL) {		/* lista libre no esta vacia */
      aux = buffer_cache->freeEnd;
      buffer_cache->buf[aux].sigFree = buffer;
    }
    else
      if (buffer_cache->freeInit == NIL)
        buffer_cache->freeInit = buffer;
      else {
        aux = buffer_cache->freeInit;
        while (buffer_cache->buf[aux].sigFree != NIL)
          aux = buffer_cache->buf[aux].sigFree;
        buffer_cache->buf[aux].sigFree = buffer;
        buffer_cache->buf[buffer].antFree = aux;
      }

    buffer_cache->freeEnd = buffer;
  }
}



/*********************************************************************
Funcin que saca el buffer de la lista libre.
**********************************************************************/
void OutFreeList(int buffer) {
int siguiente;
int anterior;
int aux;
int aux2;

  /* miramos si esta en la lista libre */
  aux = buffer_cache->freeInit;
  while ((aux != NIL) && (aux != buffer)) {
    aux2 = buffer_cache->buf[aux].sigFree;
    aux = aux2;
  }

  if (aux == buffer) {
    siguiente = buffer_cache->buf[buffer].sigFree;
    anterior = buffer_cache->buf[buffer].antFree;
    buffer_cache->buf[buffer].antFree = buffer_cache->buf[buffer].sigFree = NIL;;

    if (anterior == NIL){      			/* es el primero de la lista*/
      if (siguiente == NIL) {                      /* es el unico de la lista */
        buffer_cache->freeEnd = NIL;
        buffer_cache->freeInit = NIL;
      }
      else {
        buffer_cache->freeInit = siguiente;
        buffer_cache->buf[siguiente].antFree = NIL;
      }
   }
    else {  					/* no es el primero */
      if (siguiente == NIL)    		        /* es el ltimo */
        buffer_cache->freeEnd = anterior;
      else                             		/* cualquiera en medio de la lista */
        buffer_cache->buf[siguiente].antFree = anterior;
      buffer_cache->buf[anterior].sigFree = siguiente;
    }
  }

}

/*********************************************************************
Funcin que quita el buffer apuntado por buffer de la cola hash.
**********************************************************************/
void OutHash(int buffer) {
int siguiente;
int anterior;
int pos;
int aux;
int aux2;

  /* comprobamos que realmente esta en la cola hash */
  if (buffer_cache->buf[buffer].bloque != NIL) {
    pos = buffer_cache->buf[buffer].bloque % MAXHASH;
    aux =  buffer_cache->hash[pos];

    while (( aux != NIL) &&((buffer_cache->buf[buffer].disp != buffer_cache->buf[aux].disp) || (buffer_cache->buf[buffer].inodo != buffer_cache->buf[aux].inodo) || (buffer_cache->buf[buffer].bloque != buffer_cache->buf[aux].bloque))) {
      aux2 = buffer_cache->buf[aux].sigHash;
      aux = aux2;
    }

    if (aux != NIL) {
      siguiente = buffer_cache->buf[buffer].sigHash;
      anterior = buffer_cache->buf[buffer].antHash;
      buffer_cache->buf[buffer].validos = FALSE;

      if (anterior == NIL){     			 /* es el primero de la lista*/
        if (siguiente == NIL)                        /* es el unico de la lista */
          buffer_cache->hash[pos] = NIL;
        else {
          buffer_cache->hash[pos] = siguiente;
          buffer_cache->buf[siguiente].antHash = NIL;
        }
      }
      else {  					/* no es el primero */
        if (siguiente == NIL);             		/* es el ltimo */
     
        else                            		 /* cualquiera en medio de la lista */
          buffer_cache->buf[siguiente].antHash = anterior;
        buffer_cache->buf[anterior].sigHash = siguiente;
      }
    }
  } /* fin bloque != NIL */
}

/*********************************************************************
Funcin que pone el buffer en su correspondiente cola Hash. Se inserta
al principio de la cola.
**********************************************************************/
void PutHash(int buffer, unsigned long int bloque) {
int pos;
int siguiente;

  pos = bloque % MAXHASH;
  buffer_cache->buf[buffer].bloque = bloque;
  
  if (buffer_cache->hash[pos] == NIL) {
    buffer_cache->hash[pos] = buffer;
    buffer_cache->buf[buffer].sigHash = NIL;
    buffer_cache->buf[buffer].antHash = NIL;
  }
  else {                		 /* lo insertamos al principio */
    siguiente = buffer_cache->hash[pos];
    buffer_cache->buf[buffer].sigHash = siguiente;
    buffer_cache->buf[buffer].antHash = NIL;
    buffer_cache->hash[pos] = buffer;
    buffer_cache->buf[siguiente].antHash = buffer;
  }
}

/*********************************************************************
Funcin que devuelve un puntero a un buffer bloqueado que ya puede
ser usado. Nos podemos encontrar 5 casos:
1. El bloque al que accedemos se encuentra en la cola hash
2. El bloque no est en el buffer
3. El bloque no est en el buffer y la lista libre est vaca
4. Escritura retardada.
5. Accedemos a un bloque en el buffer cach bloqueado
Vamos a poner un semaforo para que solo un proceso pueda acceder a esta
region del cdigo en cada instante.
*********************************************************************/
int getblk(unsigned int disp, unsigned long int inodo, unsigned long int bloque) {
int buffer;
boolean encontrado = FALSE;	/* indica si el buffer ha sido encontrado o no */
int pos;                        /* indica la entrada en la cola hash */
int aux;

  Semaforo(FALSE);
  while (encontrado == FALSE) {
    /* miramos si est en la cola hash */
    pos = bloque % MAXHASH;
    buffer = buffer_cache->hash[pos];
    while ((buffer != NIL) && ((buffer_cache->buf[buffer].disp != disp) || (buffer_cache->buf[buffer].inodo != inodo) || (buffer_cache->buf[buffer].bloque != bloque))) {
      aux = buffer_cache->buf[buffer].sigHash;
      buffer = aux;
    }

    if (buffer != NIL) {		/* bloque en la cola hash */
      if (buffer_cache->buf[buffer].bloqueado) {		/* comprobamos si est bloqueado */
        /* insertamos el pid del proceso en la lista de dormidos */
        if(Insertar(buffer_cache->buf[buffer].dormidos, getpid()) == -1) {
          printf("ERROR al insertar en la lista de dormidos\n");
          Semaforo(TRUE);
          exit(1);
        }
        /* ponemos el proceso a dormir */
        Semaforo(TRUE);
        pause();
        Semaforo(FALSE);
        est.dormidos++;
        continue;
      }
      /* no est bloqueado */
      est.encontrados++;
      buffer_cache->buf[buffer].bloqueado = TRUE;          /* lo bloqueamos */
      OutFreeList(buffer);                                 /* lo quitamos de la lista libre */
      Semaforo(TRUE);
      return buffer;
    }

    else {				/* el bloque no est en la cola */
      if (buffer_cache->freeInit == NIL){    /* no hay buffers libres */
        /* lo insertamos en la lista de dormidos esperando cualquier buffer */
        if(Insertar(buffer_cache->pid_dormidos, getpid()) == -1) {
          printf("ERROR al insertar en la lista de dormidos\n");
          Semaforo(TRUE);
          exit(1);
        }
        Semaforo(TRUE);
        pause();                  /* lo dormimos */
        Semaforo(FALSE);
        est.dormidos++;
        continue;
      }
      /* cogemos el primero de la lista libre */
      buffer = PopFreeList();
      if (buffer_cache->buf[buffer].retardada) {       	/* marcado como escritura retardada */
	/* seal al proceso que abri el fichero para que escriba */
	if ((kill(buffer_cache->buf[buffer].pid, SIGUSR2)) == -1)
	  perror("seal escritura retardada");
	continue;
      }
      est.nuevos++;
      /* lo quitamos de la cola hash antigua */
      OutHash(buffer);
      /* lo ponemos en la nueva cola hash */
      PutHash(buffer,bloque);
      buffer_cache->buf[buffer].bloqueado = TRUE;
      Semaforo(TRUE);
      return buffer;
    }

  } /* fin encontrado */

}

/************************************************************************************************
Funcin que lee un bloque de disco a memoria. Primero comprueba si esta en el buffer cach.
Devuelve un buffer bloqueado con los datos vlidos.
************************************************************************************************/
int bread(int fd_real, unsigned int disp, unsigned long int inodo, unsigned long int bloque) {
int buffer;
long int pos;	/* indica el offset dentro del fichero */
int contador;

  buffer = getblk(disp, inodo, bloque);
  if (buffer_cache->buf[buffer].validos)	/* si los datos son validos */
    return buffer;

  /* datos no validos */
  /* inicia lectura disco */
  pos = TAMBUF * bloque;
  lseek(fd_real, pos, SEEK_SET);
  contador = read(fd_real, buffer_cache->buf[buffer].datos, TAMBUF);
  if (contador == 0)  {
    buffer_cache->buf[buffer].validos = FALSE;
    buffer_cache->buf[buffer].bloqueado = FALSE;
    return buffer;
  }

  buffer_cache->buf[buffer].validos = TRUE;
  buffer_cache->buf[buffer].bloque = bloque;
  buffer_cache->buf[buffer].inodo = inodo;
  buffer_cache->buf[buffer].disp = disp;

  return buffer;
}

/************************************************************************************************
Funcin que libera un buffer.
************************************************************************************************/
void brelse(int buffer) {
int i;

  Semaforo(FALSE);
  if (buffer_cache->buf[buffer].validos)	/* si los datos son validos */
    PutFinFreeList(buffer);			/* ponemos el buffer al final de la free list */		
  else
    PutInicioFreeList(buffer);			/* ponemos el buffer al principio de la free list */

  buffer_cache->buf[buffer].bloqueado = FALSE;

  for (i=0; i < MAXDORM; i++) {
    /* despertamos todos los procesos esperando por este buffer */
    if ((buffer_cache->buf[buffer].dormidos[i] != NIL) && (buffer_cache->buf[buffer].dormidos[i] != 0)){
      if ((kill(buffer_cache->buf[buffer].dormidos[i], SIGUSR1)) == -1)
        perror("ERROR seal kill");

      buffer_cache->buf[buffer].dormidos[i] = NIL;		/* posicion libre */
    }

    /* despertamos todos los procesos esperando por cualquier buffer*/
    if ((buffer_cache->pid_dormidos[i] != NIL) && (buffer_cache->pid_dormidos[i] != 0)){
      if ((kill(buffer_cache->pid_dormidos[i], SIGUSR1)) == -1)
        perror("ERROR seal kill");

      buffer_cache->pid_dormidos[i] = NIL;		/* posicion libre */
    }
  }

  est.liberados++;
  Semaforo(TRUE);
}
