#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>	/* memoria compartida */
#include <sys/sem.h>
#include "estructuras.h"

/*********************************************************************
Funcion que da memoria de la zona de memoria compartida al buffer
cache. Devuelve -1 si ha ocurrido algun error.
Tambien incializa el semaforo
*********************************************************************/
int Da_Memoria(void) {
int id;		/* identificador de la memoria compartida del buffer cache*/
int id2;	/* identificador de la memoria compartida para la tabla de ficheros*/
int semid;	/* identificador del semaforo */

  /* creamos la zona de memoria compartida. Se crea con permisos de lectura
  y modificacion para el usuario */
  id = shmget(LLAVE, TAM, IPC_CREAT | 0600);
  if (id == (-1))
    return(-1);

  id2 = shmget(LLAVE2, TAM, IPC_CREAT | 0600);
  if (id2 == (-1))
    return(-1);

  if ((semid = semget(LLAVE3,1,IPC_CREAT|0600)) == -1)
    perror("semaforo");

  buffer_cache = (BUFFER_CACHE *)shmat(id, 0,0);
  tabla_ficheros = (TABLA_FICHEROS *)shmat(id2, 0,0);
  /* inicializacion semaforo a abierto */
  semctl(semid, 0, SETVAL, 1);

  return 0;
}


/* Funcion que muestra le estado de la lista libre por pantalla */
void FreeList(void) {
int buffer;
int aux;


  buffer = buffer_cache->freeInit;
  while (buffer != NIL) {
    printf("|%d|->", buffer);
    aux = buffer_cache->buf[buffer].sigFree;
    buffer = aux;
  }
  printf("NIL\n");
  printf("Final lista libre: %d", buffer_cache->freeEnd);
}

/* Funcion que muestra la lista de dormidos de un buffer en particular
o la lista de dormidos esperando porcualquier buffer */
void Dormidos(unsigned long int *lista) {
int i;

  for (i = 0; i < MAXDORM; i++)
    if (lista[i] != NIL)
      printf("%ld ", lista[i]);
}

/* Funcion que muestra el valor de todos los campos de un buffer */
void Estado(int buffer) {

  printf("Disp: \t\t%d\n", buffer_cache->buf[buffer].disp);
  printf("Inodo: \t\t%ld\n", buffer_cache->buf[buffer].inodo);
  printf("Bloque: \t%ld\n", buffer_cache->buf[buffer].bloque);

  if (buffer_cache->buf[buffer].bloqueado)
    printf("Bloqueado: \tTRUE\n");
  else
    printf("Bloqueado: \tFALSE\n");

  if (buffer_cache->buf[buffer].retardada) {
    printf("Retardada: \tTRUE\n");
    printf("Proceso: \t%ld\n", buffer_cache->buf[buffer].pid);
    printf("Fichero: \t%ld\n", buffer_cache->buf[buffer].fd);
  }
  else
    printf("Retardada: \tFALSE\n");

  if (buffer_cache->buf[buffer].validos)
    printf("Validos: \tTRUE\n");
  else
    printf("Validos: \tFALSE\n");

  printf("antFree: \t%d\n",buffer_cache->buf[buffer].antFree);
  printf("sigFree: \t%d\n",buffer_cache->buf[buffer].sigFree);
  printf("antHash: \t%d\n",buffer_cache->buf[buffer].antHash);
  printf("sigHash: \t%d\n",buffer_cache->buf[buffer].sigHash);

  printf("Tamao: \t%ld\n", buffer_cache->buf[buffer].tamano);
  printf("Inicial: \t%ld\n", buffer_cache->buf[buffer].inicial);
  printf("Datos: \t%s", buffer_cache->buf[buffer].datos);

}

/***************************************************************************/
/* Funcin que muestra el menu en pantalla y devuelve un numero con la opcion
seleccionada */
int Menu(void) {
int opcion;

  printf("\n-------------------------------------------\n");
  printf("1. Ver lista libre.\n");
  printf("2. Ver lista dormidos por un buffer.\n");
  printf("3. Ver lista dormidos por cualquier buffer.\n");
  printf("4. Ver estado del semaforo.\n");
  printf("5. Ver estado de un buffer.\n");
  printf("6. \n");
  printf("7. \n");
  printf("0. Salir\n\n");

  printf("Opcion: ");
  scanf("%d",&opcion);
  printf("-------------------------------------------\n");

  return(opcion);
}


/***********************************************************************/
int main() {
int opcion;
int buffer;

  if (Da_Memoria() == -1)
    perror("No hay memoria para las estrucutras");

  do {
    opcion = Menu();
    switch (opcion) {
      /* ver lista llibre */
      case 1: FreeList();
	      break;
      /* Ver lista dormidos por un buffer */
      case 2: printf("Introduce el buffer: ");
	      scanf("%d", &buffer);
	      Dormidos(buffer_cache->buf[buffer].dormidos);
	      break;
      /* Ver lista dormidos por cualquier buffer */
      case 3:
	      break;
      case 4:
	      break;
      /* ver estado de un buffer */
      case 5: printf("Introduce el buffer: ");
	      scanf("%d", &buffer);
              Estado(buffer);
	      break;
      case 6:
	      break;
      case 7:
              break;

     default: break;
    }
  }while (opcion != 0);


  return 0;
}
