/********************************************************************
Mdulo que contiene las estructuras necesarias para la simulacin
del buffer cach.
********************************************************************/
#include "util.h"
#include <signal.h>
#include <sys/types.h>

#define MAXBUF 5	/* maximo numero de buffers */
#define MAXHASH 3	/* maximo numero de colas hash */
#define MAXDORM 10	/* maximo numero de procesos dormidos */
#define TAMBUF 10	/* tamao de los datos del buffer */
#define MAXFILES 15	/* maximo numero de ficheros */

/* memoria compartida */
#define TAM 2048	/* Tamao de la memoria compartida */
#define LLAVE 123456    /* llave para el buffer cache */
#define LLAVE2 123457	/* llave para la tabla de ficheros */
#define LLAVE3 123458	/* llave para el semadforo */

/* permisos de ficheros */
#define LECTURA 0
#define ESCRITURA 1
#define L_E 2			/* lectura escritura */

#define NIL -1


typedef struct TIPO_BUFFER{
  int inicial; 			/* primer byte valido */
  /* campos para la identificacion */
  unsigned int disp;		/* dispositivo */
  unsigned long int inodo;		/* inodo */
  unsigned long int bloque;		/* n de bloque */
  /* campos para el estado */
  boolean bloqueado;	/* */
  boolean retardada;	/* escritura retardada */
  boolean validos;	/* datos vlidos o no */
  /* punteros */
  int sigHash;		/* siguiente buffer segun la cola hash */
  int sigFree;		/* siguiente buffer en la lista libre */
  int antHash;		/* anterior buffer segun la cola hash */
  int antFree; 		/* anterior buffer en la lista libre */
  unsigned long int dormidos[MAXDORM];	/* pid de los procesos dormidos esperando por este buffer */
  unsigned long int pid;			
  int fd;                       /* pid del proceso que tiene que escribir este buffer a disco */
  unsigned char datos[TAMBUF];
  unsigned int tamano;
}BUFFER;

typedef struct TIPO_BUFFER_CACHE{
  int freeInit;	/* comienzo de la free list */
  int freeEnd;		/* final de la free list */
  int hash[MAXHASH];
  pid_t pid_dormidos[MAXDORM];	/* pid de los procesos esperando que se libere cualquier buffer */
  BUFFER buf[MAXBUF];
}BUFFER_CACHE;


/* estructura para la tabla de ficehros */
typedef struct TIPO_ENTRADA {
  long int offset;
  long int inodo;          	/* -1 indica entrada libre */
  int disp;
  boolean lectura;		/* permisos de lectura */
  boolean escritura;		/* permisos de escritura */
  long int tam;			/* tamao del fichero */
}ENTRADA;

typedef struct TIPO_TABLA_FICHEROS {
  ENTRADA entrada[MAXFILES];
}TABLA_FICHEROS;

/* variables globales a todos los mdulos */
BUFFER_CACHE *buffer_cache;
TABLA_FICHEROS *tabla_ficheros;
