/*************************************************************************************
  Autor: Oscar Martnez Valpuesta -- 4 Ingenieria Superior Informtica
  Descripcin: Modulo que crea una zona de memoria compartida e inicializa la estructura
		del buffer_cache.
*************************************************************************************/
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>	/* memoria compartida */
#include <sys/sem.h>
#include "estructuras.h"


/***********************************************************************************
Funcin que inicializa todas las estructuras del buffer cache.
 **********************************************************************************/
void Inicializar(void) {
register int i;
int siguiente;
int anterior;
register int j;


    /* para cada buffer */
    buffer_cache->freeInit = 0;
    buffer_cache->freeEnd = MAXBUF - 1;
    anterior = NIL;
    for (i = 0; i < MAXBUF;i++) {
      buffer_cache->buf[i].sigFree = i+1;
      buffer_cache->buf[i].antFree = anterior;
      anterior = i;
      /* punteros hash */
      buffer_cache->buf[i].sigHash = NIL;
      buffer_cache->buf[i].antHash = NIL;
      /* valores booleanos */
      buffer_cache->buf[i].bloqueado = FALSE;
      buffer_cache->buf[i].retardada = FALSE;
      buffer_cache->buf[i].validos = FALSE;
      buffer_cache->buf[i].bloque = NIL;
      buffer_cache->buf[i].tamano = 0;
      buffer_cache->buf[i].inicial = NIL;
      /* lista de dormidos (NIL indica posicion libre) */
      for (j = 0; j < MAXDORM; j++)
        buffer_cache->buf[i].dormidos[j] = NIL;
    }
    buffer_cache->buf[MAXBUF-1].sigFree = NIL;
    buffer_cache->buf[0].antFree = NIL;

    /* cola hash */
    for (i = 0; i < MAXHASH;i++)
      buffer_cache->hash[i] = NIL;

    /* lista de PID  dormidos */
    for (i = 0; i <MAXDORM; i++)
      buffer_cache->pid_dormidos[i] = NIL;


  /* tabla de ficheros */
  for (i = 0; i < MAXFILES; i++)
    tabla_ficheros->entrada[i].offset = -1;

}

/***********************************************************************/
int main() {
int id;		/* identificador de la memoria compartida */
int id2;
int semid;


  /* creamos la zona de memoria compartida. Se crea con permisos de lectura
  y modificacion para el usuario */
  id = shmget(LLAVE, TAM, IPC_CREAT | 0600);
  if (id == (-1))
    Error();

  /* memoria para la tabla de ficheros */
  id2 = shmget(LLAVE2, TAM, IPC_CREAT | 0600);
  if (id2 == (-1))
    Error();

 if ((semid = semget(LLAVE3,1,IPC_CREAT|0600)) == -1)
    perror("semaforo");

  buffer_cache = (BUFFER_CACHE *)shmat(id, 0,0);
  tabla_ficheros = (TABLA_FICHEROS *)shmat(id2, 0,0);
  /* inicializacion semaforos a abierto */
  if(semctl(semid, 0, SETVAL, 1) == -1)
    perror("inicializar semaforo");

  Inicializar();

  /* Pausa para que otros procesos puedan usar la zona de memoria */
  Pausa();


  /* finalizamos la zona de memoria compartida y el semaforo */
  if (shmctl(id, IPC_RMID, 0) == -1)
    Error();
  if (shmctl(id2, IPC_RMID, 0) == -1)
    Error();
  if (semctl(semid,0,IPC_RMID, 0) == -1)
    Error();

  return 0;
}
