#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>


/* funcin que muestra el n del error, el mensaje asociado y finaliza la ejecucin*/
void Error(void) {

  fprintf(stderr, "ERROR %d. ", errno);
  fprintf(stderr, "%s\n", strerror(errno));
  exit(1);
}


void menuDir(const char *directorio_name) {
DIR *ptrDirectorio;
struct dirent *directorio;
struct stat inodo;

  printf("\nCONTENIDO DEL DIRECTORIO %s\n", directorio_name);

  /* abrimos el directorio */
  ptrDirectorio = opendir(directorio_name);
  if (ptrDirectorio == NULL)
    Error();

  /* leemos el directorio */
  directorio = readdir(ptrDirectorio);
  while (directorio != NULL) {
    /* obtenemos informacin de cada elemento del directorio */
    if (stat(directorio->d_name, &inodo) == -1)
      Error();

    printf("%s\t\t", directorio->d_name);

    /* permisos */
    printf("%o", (inodo.st_mode & 0x1FF) >> 6);
    printf("%o", (inodo.st_mode & 56) >> 3);
    printf("%o\t", (inodo.st_mode & 0x007));

    printf("%d\n", inodo.st_size);

    directorio = readdir(ptrDirectorio);
  }

  /* cerramos el directorio */
  if (closedir(ptrDirectorio) == -1)
    Error();

}

void menuFile(const char *fichero) {
int opcion;
int permisos;
char s_n;
char new_name[255];

  printf("\nFICHERO REGULAR\n"); 
  printf("1. Cambiar permisos.\n");
  printf("2. Renombrar.\n");
  printf("3. Borrar.\n");
  printf("\nOpcin: ");
  fflush(NULL); 
  scanf("%d", &opcion);

  switch(opcion) {
    case 1: printf("Introduzca los nuevos permisos (nnn): ");
            scanf("%o", &permisos);
	    if (chmod(fichero, permisos) == -1)
              Error();
            break;
    case 2: printf("Introduzca el nuevo nombre: ");
	    scanf("%s", &new_name);
 	    if (rename(fichero, new_name) == -1)
              Error();
            else
               printf("Fichero renombrado.\n");
	    break;
    case 3: printf("Est seguro (s/n)?: ");
            fflush(NULL);
            scanf("%s", &s_n); 
	    if (s_n == 's') {  /* borramos el fichero */
              unlink(fichero);
              printf("Fchero borrado\n");
            }
            break;
    default: printf("Opcin incorrecta\n");
             break;
  }
}

void menuDisp(int disp) {
  
  printf("\nDISPOSITIVO %d\n", disp);

  printf("Major: %d\n", major(disp));
  printf("Minor: %d\n", minor(disp));
}


int main(int argc, char *argv[]){
struct stat inodo;

  if (argc != 2) {
    printf("N de argumentos incorrecto.\n");
    exit(1);
  }

  if (stat(argv[1], &inodo) == -1)
    Error();


  printf("Inodo: %ld\n", inodo.st_ino);
  printf("Tamao: %d\n", inodo.st_size);
  printf("File system: %d\n", inodo.st_dev);

  /* permisos */
  printf("Permisos: \n");
  printf("   Usuario: \t%o\n", (inodo.st_mode & 0x1FF) >> 6);
  printf("   Grupo: \t%o\n", (inodo.st_mode & 56) >> 3); 
  printf("   Resto: \t%o\n", (inodo.st_mode & 0x007));

  printf("Identificacin usuario propietario: %d\n", inodo.st_uid);
  printf("Identificacin grupo propietario: %d\n", inodo.st_gid);
  printf("N de hard links: %d\n", inodo.st_nlink);

  printf("Tamao del bloque: %d\n", inodo.st_blksize);
  printf("Nmero de bloques usados: %d\n", inodo.st_blocks);
  
  printf("ltimo acceso: %s\n", ctime(&inodo.st_atime));
  printf("Ultima modificacin: %s\n", ctime(&inodo.st_mtime));
  printf("Ultimo cambio: %s\n", ctime(&inodo.st_ctime)); 

  
  /* dispositivos */
  if (S_ISREG(inodo.st_mode))
    menuFile(argv[1]);
  if (S_ISDIR(inodo.st_mode))
    menuDir(argv[1]);
  if (S_ISCHR(inodo.st_mode) || S_ISBLK(inodo.st_mode))
    menuDisp(inodo.st_rdev);

  return 0;
}
